/*   
 *   $Id: db.c,v 1.32 2002/07/12 19:30:32 ondra Exp $
 * 
 *   Holiday Debts manager for PalmOS.
 *   Copyright (C) 2000,2001 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *   You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "debts.h"
#include "db.h"
#include "debtsRsc.h"

static DmOpenRef gDB=NULL;

/***********************************************************************
 *
 * FUNCTION:     db_set_backup
 * 
 * DESCRIPTION:  Set backup bit on database
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
db_set_backup(void)
{
    UInt16 cardNo;
    LocalID dbID;
    UInt16 attribute;
    
    if (DmOpenDatabaseInfo(gDB, &dbID, NULL, NULL, &cardNo, NULL))
	return;
    if (DmDatabaseInfo(cardNo, dbID, NULL,&attribute,NULL,NULL,
		       NULL,NULL,NULL,NULL,
		       NULL,NULL,NULL))
	return;
    attribute |= dmHdrAttrBackup;
    DmSetDatabaseInfo(cardNo,dbID,NULL,&attribute,NULL,
		      NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
}

/***********************************************************************
 *
 * FUNCTION:     db_get_prefs
 * 
 * DESCRIPTION:  Return database preferences (default currency, etc.)
 *
 * PARAMETERS:   None
 *
 * RETURN:       dbPrefs
 *      
 ***********************************************************************/
dbPrefs db_get_prefs(void)
{
    UInt16 cardNo;
    LocalID dbID;
    LocalID appInfoID;
    dbPrefs result,*aprefs;

    StrCopy(result.lcurrency,defaultCurrency);
    result.curidx = -1;
    result.fuser = -1;
    result.firstline = 0;

    if (DmOpenDatabaseInfo(gDB, &dbID, NULL, NULL, &cardNo, NULL))
	return result;
    if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,
		       NULL,NULL,&appInfoID, NULL, NULL, NULL))
	return result;
    if (appInfoID == NULL) 
	return result;
    
    aprefs = MemLocalIDToLockedPtr(appInfoID,cardNo);
    result = *aprefs;
    MemPtrUnlock(aprefs);

    return result;
}

/***********************************************************************
 *
 * FUNCTION:     db_save_prefs
 * 
 * DESCRIPTION:  Save database preferences
 *
 * PARAMETERS:   dbPrefs
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void 
db_save_prefs(dbPrefs prefs)
{
    UInt16 cardNo;
    LocalID dbID;
    LocalID appInfoID;
    dbPrefs *aprefs;
    MemHandle h;

    if (DmOpenDatabaseInfo(gDB, &dbID, NULL, NULL, &cardNo, NULL))
	return;
    if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,
		       NULL,NULL,&appInfoID, NULL, NULL, NULL))
	return;
    if (appInfoID == NULL) {
	h = DmNewHandle(gDB,sizeof(prefs));
	appInfoID = MemHandleToLocalID(h);
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL,
			  NULL, NULL, NULL, &appInfoID, NULL, NULL,
			  NULL);
    }
    aprefs = MemLocalIDToLockedPtr(appInfoID, cardNo);
    DmWrite(aprefs,0,&prefs,sizeof(prefs));
    MemPtrUnlock(aprefs);
}


/***********************************************************************
 *
 * FUNCTION:     db_free_itemlist
 * 
 * DESCRIPTION:  Free memory associated with itemlist
 *
 * PARAMETERS:   itemlist
 *
 * RETURN:       None
 *      
 ***********************************************************************/
void
db_free_itemlist(dbItemList *arr)
{
    MemPtrFree(arr);
}

/***********************************************************************
 *
 * FUNCTION:     db_get_itemlist
 * 
 * DESCRIPTION:  Return list of items (the payments) filtered by user,
 *               if user==-1, return all
 *
 * PARAMETERS:   user
 *
 * RETURN:       dbItemList
 *      
 ***********************************************************************/
dbItemList *
db_get_itemlist(Int16 user)
{
    Int16 totalcount,i,j,k;
    dbItemList *result;
    itemType *item;

    totalcount = db_num_items();

    result = MemPtrNew(sizeof(*result) + sizeof(result->seq[0]) * totalcount);
    if (user == -1) {
	for (i=0;i<totalcount;i++) 
	    result->seq[i] = i;
	result->nitems = totalcount;
    } else {
	for (i=0,j=0;i<totalcount;i++) {
	    item = db_get_item(i);
	    if (item->payer == user) 
		result->seq[j++] = i;
	    else 
		for (k=0;k<item->npayees;k++)
		    if (item->pmts[k].id == user)
			result->seq[j++] = i;
	    db_free_item(item);
	}
	result->nitems = j;
    }
    return result;
}

/***********************************************************************
 *
 * FUNCTION:     db_compute_balance
 * 
 * DESCRIPTION:  Compute the total balance of the user
 *
 * PARAMETERS:   user
 *               curidx - currency, in which the result should
 *                        be given 
 *
 * RETURN:       payment, if paid by user - cost allocated to this user
 *      
 ***********************************************************************/
double
db_compute_balance(Int16 user,Int16 curidx)
{
    itemType *item;
    Int16 i,totalcount;
    double balance = 0.0;

    if (user == -1)
	return 0.0;

    totalcount = db_num_items();
    for (i=0;i < totalcount;i++) {
	item = db_get_item(i);
	balance += get_item_costs(item,user,USERBALANCE,curidx);
	db_free_item(item);
    }
    return balance;
}

/***********************************************************************
 *
 * FUNCTION:     db_compute_cost
 * 
 * DESCRIPTION:  Compute the total cost for the user 
 *
 * PARAMETERS:   user - user for which it should be computed 
 *                      or -1 for all
 *               curidx - currency, in which the result should
 *                        be given 
 *
 * RETURN:       
 *      
 ***********************************************************************/
double
db_compute_cost(Int16 user,Int16 curidx)
{
    itemType *item;
    Int16 i,totalcount;
    double cost = 0.0;

    totalcount = db_num_items();
    for (i=0;i < totalcount;i++) {
	item = db_get_item(i);
	cost += get_item_costs(item,user,USERCOST,curidx);
	db_free_item(item);
    }
    return cost;
}

/***********************************************************************
 *
 * FUNCTION:     db_compute_cost
 * 
 * DESCRIPTION:  Compute the total cost for the user 
 *
 * PARAMETERS:   user - user for which it should be computed 
 *                      or -1 for all
 *               curidx - currency, in which the result should
 *                        be given 
 *
 * RETURN:       
 *      
 ***********************************************************************/
double
db_compute_payment(Int16 user,Int16 curidx)
{
    itemType *item;
    Int16 i,totalcount;
    double payment = 0.0;

    totalcount = db_num_items();
    for (i=0;i < totalcount;i++) {
	item = db_get_item(i);
	payment += get_item_costs(item,user,USERPAYMENT,curidx);
	db_free_item(item);
    }
    return payment;
}

Int16
compare_int(Int16 a,Int16 b)
{
    if (a > b)
	return 1;
    else if (a < b)
	return -1;
    return 0;
}

static Int16
db_compare_record(dbRecord *rec1,dbRecord *rec2,
		  Int16 unused1, SortRecordInfoPtr unused2,
		  SortRecordInfoPtr unsed3,MemHandle appInfoH)
{
    if (rec1->type > rec2->type)
	return 1;
    else if (rec1->type < rec2->type)
	return -1;
    
    if (rec1->type == db_person) 
	return compare_int(rec1->u.person.id, rec2->u.person.id);

    if (rec1->type == db_currency)
	return compare_int(rec1->u.currency.id, rec2->u.currency.id);
    
    if (rec1->type == db_description)
	return StrCompare(rec1->u.description, rec2->u.description);
    
    if (rec1->type == db_item) {
	if (rec1->u.item.date.year != rec2->u.item.date.year)
	    return compare_int(rec1->u.item.date.year,rec2->u.item.date.year);
	if (rec1->u.item.date.month != rec2->u.item.date.month)
	    return compare_int(rec1->u.item.date.month,rec2->u.item.date.month);
	if (rec1->u.item.date.day != rec2->u.item.date.day)
	    return compare_int(rec1->u.item.date.day,rec2->u.item.date.day);
	/* This is just a trick, unused get set only for new items,
         * so that they get inserted last */
	return compare_int(rec1->u.item.unused,rec2->u.item.unused);
    }

    return 0;
}

/***********************************************************************
 *
 * FUNCTION:     db_first_item
 * 
 * DESCRIPTION:  Finds the index of the first item(payement...) in
 *               the database ('items' are sorted in the end of DB)
 *
 * PARAMETERS:   None
 *
 * RETURN:       Index of the first record of type db_item
 *      
 ***********************************************************************/
static UInt16 
db_first_item(void)
{
    UInt16 i;
    dbRecord *record;
    MemHandle handle;
    Boolean foundit = false;
    
    for (i=0;i<DmNumRecords(gDB);i++) {
	handle = DmQueryRecord(gDB,i);
	record = MemHandleLock(handle);
	foundit = record->type == db_item;
	MemHandleUnlock(handle);
	if (foundit)
	    break;
    }
    return i;
}

Boolean
db_delete_item(Int16 pos)
{
    return DmRemoveRecord(gDB,pos + db_first_item());
}

Boolean
db_add_item(itemRec *item)
{
    dbRecord *record,*tmprecord;
    MemHandle tmphandle;
    UInt16 rnum;
    Int16 size,i;
    
    db_set_backup();

    size = sizeof(*record) + sizeof(record->u.item.pmts[0])*item->npayees;
    record = MemPtrNew(size);

    record->type = db_item;
    
    record->u.item.date = item->date;
    record->u.item.amount = item->amount;
    record->u.item.payer = item->payer;
    record->u.item.transfer = item->transfer;
    record->u.item.currency = item->currency;
    StrCopy(record->u.item.description,item->description);
    record->u.item.npayees = item->npayees;
    for (i=0;i<item->npayees;i++)
	record->u.item.pmts[i] = item->pmts[i];

    /* Insert it as a last item of that day, this is a trick */
    record->u.item.unused = 1;
    rnum = DmFindSortPosition(gDB,record,0,
			      (DmComparF *) db_compare_record,0);
    record->u.item.unused = 0;

    tmphandle = DmNewRecord(gDB,&rnum,size);
    tmprecord = MemHandleLock(tmphandle);
    DmWrite(tmprecord,0,record,size);
    MemHandleUnlock(tmphandle);
    DmReleaseRecord(gDB,rnum,true);
    
    MemPtrFree(record);

    return true;
}

void
db_free_item(itemType *item)
{
    MemPtrFree(item);
}

itemType *
db_get_item(Int16 rnum)
{
    dbRecord *record;
    MemHandle handle;
    itemType *item;
    Int16 i;
    
    handle = DmQueryRecord(gDB, rnum + db_first_item());
    record = MemHandleLock(handle);
    
    item = MemPtrNew(sizeof(*item) 
		     + sizeof(item->pmts[0])*record->u.item.npayees);

    item->date = record->u.item.date;
    item->amount = record->u.item.amount;
    item->payer = record->u.item.payer;
    item->transfer = record->u.item.transfer;
    item->currency = record->u.item.currency;
    StrCopy(item->description,record->u.item.description);
    item->npayees = record->u.item.npayees;
    for (i=0;i<record->u.item.npayees;i++)
	item->pmts[i] = record->u.item.pmts[i];
    
    MemHandleUnlock(handle);

    return item;
}

UInt16
db_num_items(void)
{
    return DmNumRecords(gDB) - db_first_item();
}

Boolean
db_currency_inuse(Int16 id)
{
    UInt16 rnum;
    dbRecord *record;
    UInt16 numitems = DmNumRecords(gDB);
    Boolean foundit = false;
    MemHandle handle;

    for (rnum=0;rnum < numitems;rnum++) {
	handle = DmQueryRecord(gDB, rnum);
	record = MemHandleLock(handle);
	foundit = (record->type==db_item && record->u.item.currency==id);
	MemHandleUnlock(handle);
	if (foundit)
	    break;
    }
    return foundit;
}

Boolean
db_delete_currency(Int16 id)
{
    dbRecord srcrecord;
    UInt16 rnum;
    MemHandle handle;
    dbRecord *record;
    Boolean foundit = false;

    db_set_backup();

    srcrecord.type = db_currency;
    srcrecord.u.currency.id = id;
    rnum = DmFindSortPosition(gDB, &srcrecord, 0,
			      (DmComparF *) db_compare_record,0);
    if (rnum > 0) {
	handle = DmQueryRecord(gDB, rnum-1);
	record = MemHandleLock(handle);
	foundit = (record->type==db_currency && record->u.currency.id == id);
	MemHandleUnlock(handle);
	if (foundit) 
	    DmRemoveRecord(gDB, rnum-1);
    }
    return foundit;
}

Boolean
db_delete_person(Int16 id)
{
    dbRecord srcrecord;
    UInt16 rnum;
    MemHandle handle;
    dbRecord *record;
    Boolean foundit = false;

    db_set_backup();

    srcrecord.type = db_person;
    srcrecord.u.person.id = id;
    rnum = DmFindSortPosition(gDB, &srcrecord, 0,
			      (DmComparF *) db_compare_record,0);
    if (rnum > 0) {
	handle = DmQueryRecord(gDB, rnum-1);
	record = MemHandleLock(handle);
	foundit = (record->type == db_person) && (record->u.person.id == id);
	MemHandleUnlock(handle);
	if (foundit) 
	    DmRemoveRecord(gDB, rnum-1);
    }
    return foundit;
}

Boolean
db_delete_description(Int16 seq)
{
    dbRecord *record;
    Int16 i,maxrec,pcount;
    MemHandle rechandle;
    
    maxrec = DmNumRecords(gDB);
    for (i=0,pcount=0;i<maxrec;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_description) {
	    if (pcount == seq) {
		MemHandleUnlock(rechandle);
		DmRemoveRecord(gDB,i);
		return true;
	    }
	    pcount++;
	}
	MemHandleUnlock(rechandle);
    }
    return false;
}

char **
db_get_descriptions(void)
{
    char **result;
    dbRecord *record;
    Int16 i,j,maxrec,pcount;
    MemHandle rechandle;
    
    maxrec = DmNumRecords(gDB);
    for (i=0,pcount=0;i<maxrec;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_description)
	    pcount++;
	MemHandleUnlock(rechandle);
    }

    result = MemPtrNew(sizeof(*result)*(pcount+1));

    for (i=0,j=0;j<pcount;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_description) {
	    result[j] = MemPtrNew(MAX_DESCRIPTION+1);
	    StrCopy(result[j],record->u.description);
	    j++;
	}
	MemHandleUnlock(rechandle);
    }
    result[pcount] = NULL;

    return result;
}

void
db_free_descriptions(char **descriptions)
{
    Int16 i;

    for (i=0;descriptions[i];i++)
	MemPtrFree(descriptions[i]);
    MemPtrFree(descriptions);
}

void
db_add_description(char *description)
{
    MemHandle tmphandle;
    dbRecord *record,*tmprecord;
    UInt16 rnum;
    Int16 size;

    db_set_backup();
    
    size = sizeof(*record);
    record = MemPtrNew(size);
    record->type = db_description;
    StrCopy(record->u.description, description);
    
    rnum = DmFindSortPosition(gDB,record,0,
			      (DmComparF *) db_compare_record,0);
    if (rnum > 0) {
	Boolean foundit;

	tmphandle = DmQueryRecord(gDB,rnum - 1);
	tmprecord = MemHandleLock(tmphandle);
	foundit = (tmprecord->type == record->type) && \
	    (StrCompare(tmprecord->u.description,record->u.description)==0);
	MemHandleUnlock(tmphandle);

	if (foundit) {
	    MemPtrFree(record);
	    return;
	}
    }
	
    tmphandle = DmNewRecord(gDB,&rnum,size);
    tmprecord = MemHandleLock(tmphandle);
    DmWrite(tmprecord,0,record,size);
    MemHandleUnlock(tmphandle);
    DmReleaseRecord(gDB,rnum,true);

    MemPtrFree(record);
}

void
db_add_person(Int16 id,char *name,double coef)
{
    MemHandle tmphandle;
    dbRecord *record,*tmprecord;
    UInt16 rnum;
    Int16 size;

    db_set_backup();

    size = sizeof(*record)+StrLen(name)+1;
    record = MemPtrNew(size);
    record->type = db_person;
    record->u.person.id = id;
    record->u.person.coef = coef;
    StrCopy(record->u.person.name,name);
    
    rnum = DmFindSortPosition(gDB,record,0,
			      (DmComparF *) db_compare_record,0);
    if (rnum > 0) {
	Boolean foundit;

	tmphandle = DmQueryRecord(gDB,rnum - 1);
	tmprecord = MemHandleLock(tmphandle);
	foundit = (tmprecord->type == record->type) && \
	    (tmprecord->u.person.id == record->u.person.id);
	MemHandleUnlock(tmphandle);
	if (foundit) {
	    tmphandle = DmGetRecord(gDB,rnum - 1);
	    MemHandleResize(tmphandle,size);
	    tmprecord = MemHandleLock(tmphandle);
	    DmWrite(tmprecord,0,record,size);
	    MemHandleUnlock(tmphandle);
	    DmReleaseRecord(gDB,rnum - 1,true);

	    MemPtrFree(record);
	    return;
	}
    }
    tmphandle = DmNewRecord(gDB,&rnum,size);
    tmprecord = MemHandleLock(tmphandle);
    DmWrite(tmprecord,0,record,size);
    MemHandleUnlock(tmphandle);
    DmReleaseRecord(gDB,rnum,true);

    MemPtrFree(record);
}

currencyListType *
db_get_currency_list(char *lcurrency)
{
    dbRecord *record;
    Int16 i,j,maxrec,pcount;
    MemHandle rechandle;
    currencyListType *result;
    
    maxrec = DmNumRecords(gDB);
    for (i=0,pcount=0;i<maxrec;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_currency)
	    pcount++;
	MemHandleUnlock(rechandle);
    }
    pcount++;

    result = MemPtrNew(sizeof(*result));
    result->list = MemPtrNew(sizeof(result->list[0]) * pcount);

    result->list[0] = MemPtrNew(sizeof(*(result->list[0])));
    StrCopy(result->list[0]->name,lcurrency);
    result->list[0]->id = 0;
    result->list[0]->rate = 1.0;

    for (i=0,j=1;j<pcount;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_currency) {
	    result->list[j] = MemPtrNew(sizeof(*(result->list[0])));
	    *(result->list[j]) = record->u.currency;
	    j++;
	}
	MemHandleUnlock(rechandle);
    }
    result->nitems = j;

    return result;
}

void
db_add_currency(Int16 id,char *name, double rate)
{
    MemHandle tmphandle;
    dbRecord *record,*tmprecord;
    UInt16 rnum;
    Int16 size;

    db_set_backup();

    size = sizeof(*record)+StrLen(name)+1;
    record = MemPtrNew(size);
    record->type = db_currency;
    record->u.currency.id = id;
    record->u.currency.rate = rate;
    StrCopy(record->u.currency.name,name);
    
    rnum = DmFindSortPosition(gDB,record,0,
			      (DmComparF *) db_compare_record,0);
    if (rnum > 0) {
	Boolean foundit;

	tmphandle = DmQueryRecord(gDB,rnum - 1);
	tmprecord = MemHandleLock(tmphandle);
	foundit = (tmprecord->type == record->type) && \
	    (tmprecord->u.currency.id == record->u.currency.id);
	MemHandleUnlock(tmphandle);
	if (foundit) {
	    tmphandle = DmGetRecord(gDB,rnum - 1);
	    MemHandleResize(tmphandle,size);
	    tmprecord = MemHandleLock(tmphandle);
	    DmWrite(tmprecord,0,record,size);
	    MemHandleUnlock(tmphandle);
	    DmReleaseRecord(gDB,rnum - 1,true);

	    MemPtrFree(record);
	    return;
	}
    }
    tmphandle = DmNewRecord(gDB,&rnum,size);
    tmprecord = MemHandleLock(tmphandle);
    DmWrite(tmprecord,0,record,size);
    MemHandleUnlock(tmphandle);
    DmReleaseRecord(gDB,rnum,true);

    MemPtrFree(record);
}

void
db_free_peoplelist(dbPeopleList *list)
{
    Int16 i;

    if (list->people != NULL)
	MemPtrFree(list->people);
    for (i=0;i<list->nitems;i++)
	MemPtrFree(list->names[i]);
    MemPtrFree(list);
}

dbPeopleList *
db_get_peoplelist(void)
{
    dbRecord *record;
    Int16 i,j,maxrec,pcount;
    MemHandle rechandle;
    dbPeopleList *result;
    
    maxrec = DmNumRecords(gDB);
    for (i=0,pcount=0;i<maxrec;i++) {
	rechandle = DmQueryRecord(gDB,i);
	record = MemHandleLock(rechandle);
	if (record->type == db_person)
	    pcount++;
	MemHandleUnlock(rechandle);
    }
    result = MemPtrNew(sizeof(*result) + pcount*sizeof(result->names[0]));
    result->nitems = pcount;
    if (pcount) {
	result->people = MemPtrNew(sizeof(result->people[0]) * pcount);
	for (i=0,j=0;j<pcount;i++) {
	    rechandle = DmQueryRecord(gDB,i);
	    record = MemHandleLock(rechandle);
	    if (record->type == db_person) {
		result->people[j].id = record->u.person.id;
		result->people[j].coef = record->u.person.coef;
		result->names[j] = MemPtrNew(StrLen(record->u.person.name)+1);
		StrCopy(result->names[j],record->u.person.name);
		j++;
	    }
	    MemHandleUnlock(rechandle);
	}
    } else {
	result->people = NULL;
    }

    return result;
}

void
db_get_name(char *name)
{
    UInt16 cardno;
    LocalID dbid;
    
    DmOpenDatabaseInfo(gDB,&dbid,NULL,NULL,&cardno,NULL);
    DmDatabaseInfo(cardno,dbid,name,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
		   NULL,NULL,NULL);
}

void
db_copy_template(LocalID dstid,UInt16 dstcardno,LocalID srcid,UInt16 srccardno)
{
    DmOpenRef dstdb, srcdb;
    UInt16 i,recnum;
    UInt16 maxrec = dmMaxRecordIndex;
    MemHandle srch, dsth;
    dbRecord *srcrec,*dstrec;
    dbPrefs *srcprefs,*dstprefs,tmpprefs;
    LocalID dstinfoid,srcinfoid;

    srcdb = DmOpenDatabase(srccardno,srcid,dmModeReadOnly);
    dstdb = DmOpenDatabase(dstcardno,dstid,dmModeReadWrite);

    /* Copy people and currencys */
    recnum = DmNumRecords(srcdb);
    for (i=0;i<recnum;i++) {
	srch = DmQueryRecord(srcdb,i);
	srcrec = MemHandleLock(srch);
	if (srcrec->type == db_person || srcrec->type == db_description
	    || srcrec->type == db_currency) {
	    maxrec = dmMaxRecordIndex;
	    dsth = DmNewRecord(dstdb,&maxrec,MemHandleSize(srch));
	    dstrec = MemHandleLock(dsth);
	    
	    DmWrite(dstrec,0,srcrec,MemHandleSize(srch));
	    
	    MemHandleUnlock(dsth);
	    DmReleaseRecord(dstdb,maxrec,true);
	}
	MemHandleUnlock(srch);
    }
    /* Copy DB application block */
    dsth = DmNewHandle(dstdb,sizeof(*dstprefs));
    dstinfoid = MemHandleToLocalID(dsth);
    DmSetDatabaseInfo(dstcardno, dstid, NULL, NULL, NULL, NULL,
		      NULL, NULL, NULL, &dstinfoid, NULL, NULL,
		      NULL);
    dstprefs = MemLocalIDToLockedPtr(dstinfoid,dstcardno);
    
    DmDatabaseInfo(srccardno, srcid, NULL, NULL, NULL, NULL, NULL,
		   NULL,NULL,&srcinfoid, NULL, NULL, NULL);
    srcprefs = MemLocalIDToLockedPtr(srcinfoid,srccardno);

    tmpprefs = *srcprefs;
    tmpprefs.fuser = -1;
    tmpprefs.curidx = -1;
    DmWrite(dstprefs,0,&tmpprefs,sizeof(*dstprefs));
    
    MemPtrUnlock(srcprefs);
    MemPtrUnlock(dstprefs);
    
    DmCloseDatabase(srcdb);
    DmCloseDatabase(dstdb);
}

void
db_convert_v3(UInt16 cardno,LocalID dbid)
{
    LocalID appInfoID;
    dbPrefsV3 *aprefs;
    dbPrefs newprefs;
    UInt16 version;
    MemHandle h;

    if (DmDatabaseInfo(cardno, dbid, NULL, NULL, &version, NULL, NULL,
		       NULL,NULL,&appInfoID, NULL, NULL, NULL)) 
	return;
    if (version != 3)
	return;

    if (appInfoID == NULL) 
	return;
    
    aprefs = MemLocalIDToLockedPtr(appInfoID,cardno);
    newprefs.curidx = aprefs->curidx;
    newprefs.fuser = aprefs->fuser;
    newprefs.firstline = aprefs->firstline;
    StrCopy(newprefs.lcurrency,aprefs->lcurrency);
    newprefs.indirectcur = false;
    newprefs.costMode = TOTALCOST;

    h = MemPtrRecoverHandle(aprefs);
    MemPtrUnlock(aprefs);

    version = DBVERSION;
    appInfoID = 0;
    DmSetDatabaseInfo(cardno, dbid, NULL, NULL, &version, NULL,
		      NULL, NULL, NULL, &appInfoID, NULL, NULL,
		      NULL);
    MemHandleFree(h);
    db_save_prefs(newprefs);
}

void
db_convert_v4(UInt16 cardno,LocalID dbid)
{
    LocalID appInfoID;
    dbPrefsV4 *aprefs;
    dbPrefs newprefs;
    UInt16 version;
    MemHandle h;

    if (DmDatabaseInfo(cardno, dbid, NULL, NULL, &version, NULL, NULL,
		       NULL,NULL,&appInfoID, NULL, NULL, NULL)) 
	return;
    if (version != 4)
	return;

    if (appInfoID == NULL) 
	return;
    
    aprefs = MemLocalIDToLockedPtr(appInfoID,cardno);
    newprefs.curidx = aprefs->curidx;
    newprefs.fuser = aprefs->fuser;
    newprefs.firstline = aprefs->firstline;
    newprefs.indirectcur = aprefs->indirectcur;
    StrCopy(newprefs.lcurrency,aprefs->lcurrency);

    newprefs.costMode = TOTALCOST;

    h = MemPtrRecoverHandle(aprefs);
    MemPtrUnlock(aprefs);

    version = DBVERSION;
    appInfoID = 0;
    DmSetDatabaseInfo(cardno, dbid, NULL, NULL, &version, NULL,
		      NULL, NULL, NULL, &appInfoID, NULL, NULL,
		      NULL);
    MemHandleFree(h);
    db_save_prefs(newprefs);
}

Boolean
db_check_database(LocalID dbid,UInt16 cardno)
{
    DmSearchStateType state;
    UInt16 tmpcard;
    LocalID tmpid;
    Err err;
    UInt16 version;
    
    err = DmGetNextDatabaseByTypeCreator(true, &state,
					 DBTYPE,APP_ID,false,
					 &tmpcard, &tmpid);
    while (!err && tmpid) {
	if (dbid == tmpid && cardno == tmpcard) {
	    DmDatabaseInfo(cardno,dbid,
			   NULL, /* name */
			   NULL, /* attrib */
			   &version, /* version */
			   NULL, /* crDate */
			   NULL, /* modDate */
			   NULL, /* bckUpDate */
			   NULL, /* modNum */
			   NULL, /* appinfoID */
			   NULL, /* sortInfoID */
			   NULL, /* Type */
			   NULL); /* Creator */

	    if (version == DBVERSION || version == 3 || version == 4)
		return true;
		
	    FrmAlert(altBadDBVersion);
	    return false;
	}
	err = DmGetNextDatabaseByTypeCreator(false,&state,
					     DBTYPE,APP_ID,false,
					     &tmpcard,&tmpid);
    }
    return false;
}

Boolean
db_open_by_id(LocalID dbid,UInt16 cardno)
{
    /* Check if the DB exists */
    if (!db_check_database(dbid,cardno))
	return false;
    gDB = NULL;

    gDB=DmOpenDatabase(cardno,dbid,dmModeReadWrite);
    if (gDB) {
	db_convert_v3(cardno,dbid);
	db_convert_v4(cardno,dbid);
	return true;
    }
    return false;
}

Boolean
db_create_new(char *name,LocalID *id,UInt16 *cardno)
{
    Err err;
    LocalID dbid;
    UInt16 version;
	
    dbid=DmFindDatabase(CARDNO,name);	
    if (dbid) {  /* Database exists */
	return false;
    }
	
    /* Database doesn't exist or old version */
    err=DmCreateDatabase(CARDNO,name,APP_ID,DBTYPE,false);
    if (err)
	return err;
	
    dbid=DmFindDatabase(CARDNO,name);
	
    version=DBVERSION;
    DmSetDatabaseInfo(CARDNO,dbid,
		      NULL, /* name */
		      NULL, /* attrib */
		      &version, /* version */
		      NULL, /* crDate */
		      NULL, /* modDate */
		      NULL, /* bckUpDate */
		      NULL, /* modNum */
		      NULL, /* appinfoID */
		      NULL, /* sortInfoID */
		      NULL, /* Type */
		      NULL); /* Creator */

    *id = dbid;
    *cardno = CARDNO;
    return true;
}


Err
db_close()
{
    Err err;

    if (gDB == NULL)
	return 0;

    err = DmCloseDatabase(gDB);
    gDB = NULL;

    return err;
}

Err
db_delete_database(LocalID id,UInt16 cardno)
{
    return DmDeleteDatabase(cardno,id);
}
