/*   
 *   $Id: debts.c,v 1.31 2002/03/09 13:02:10 ondra Exp $
 * 
 *   Holiday Debts manager for PalmOS.
 *   Copyright (C) 2000,2001 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *   You can contact me at 'ondrap@penguin.cz'.
*/

/* Main code for HolidayDebts */

#include <PalmOS.h>

#include "debtsRsc.h"
#include "item.h"
#include "people.h"
#include "db.h"
#include "float.h"
#include "debts.h"
#include "main.h"
#include "dbmgr.h"
#include "currency.h"
#include "description.h"

Boolean palmOS31=false;

DateFormatType dateFormat;
DateFormatType longDateFormat;
char defaultCurrency[MAX_CURRENCY_NAME+1];
Boolean activeBalance;

void * 
GetObjectPtr(UInt16 objectID)
{
    FormPtr frm;

    frm=FrmGetActiveForm();
    return (FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,objectID)));
}

/***********************************************************************
 *
 * FUNCTION:     get_item_costs
 * 
 * DESCRIPTION:  Returns appropriate sum of money
 *               with respect to selected user and requested type
 *               of information
 *
 * PARAMETERS:   item - selected item
 *               person - person for which to compute the amount
 *               mode - what should we compute
 *                 - total amount of item (TOTALCOST)
 *                 - costs for this particular user (USERCOST)
 *                   (ignoring transfers)
 *                 - how does it affect the balance (USERBALANCE)
 *                 - how much did the user pay (USERPAYMENT)
 *                 - costs (not ignoring transfers - INTERCOST)
 *               curidx - the currency in which we should compute
 *
 * RETURN:       Requested amount
 *      
 ***********************************************************************/
double
get_item_costs(itemType *item,Int16 person,CostMode mode,Int16 curidx)
{
    double amount;
    double sum;
    double coef;
    Int16 i;

    amount = 0.0;
    if (mode == TOTALCOST)
	amount = item->amount;
    else if (mode == USERPAYMENT) {
	if (item->payer == person)
	    amount = item->amount;
    } else if (mode == USERCOST) {
	if (!item->transfer)
	    amount = get_item_costs(item,person,INTERCOST,-1);
    } else if (mode == INTERCOST) {
	if (person == -1)
	    amount = item->amount;
	else {
	    coef = sum = 0.0;
	    for (i=0;i<item->npayees;i++) {
		sum += item->pmts[i].coef;
		if (item->pmts[i].id == person)
		    coef = item->pmts[i].coef;
	    }
	    amount = (item->amount/sum) * coef;
	}
    } else { /* (mode == USERBALANCE) */
	amount = get_item_costs(item,person,USERPAYMENT,-1) - 
	    get_item_costs(item,person,INTERCOST,-1);
	if (!activeBalance)
	    amount = -amount;
    }
    /* Convert result to requested currency */
    if (curidx != -1 && currencyList->list[curidx]->id != item->currency) {
	/* Convert to base currency */
	for (i=0;i < currencyList->nitems;i++)
	    if (currencyList->list[i]->id == item->currency)
		break;
	amount *= currencyList->list[i]->rate;
	/* Convert to requested currency */
	amount /= currencyList->list[curidx]->rate;
    }
    return amount;
}

void
debts_save_dbmgr(void)
{
    PrefData data;

    data.start = pref_dbmgr;
    data.activeBalance = activeBalance;
    StrCopy(data.lcurrency,defaultCurrency);
    PrefSetAppPreferences(APP_ID,PREF_DEFAULT,PREF_VERSION,
			   &data,sizeof(data),false);
}

void
debts_save_main(LocalID dbid,UInt16 cardno)
{
    PrefData data;
    
    data.start = pref_main;
    data.dbid = dbid;
    data.cardno = cardno;
    data.activeBalance = activeBalance;
    StrCopy(data.lcurrency,defaultCurrency);
    PrefSetAppPreferences(APP_ID,PREF_DEFAULT,PREF_VERSION,
			   &data,sizeof(data),false);
}

static Boolean
calc_rom_greater(UInt16 major, UInt16 minor)
{
	UInt32        romVersion;
	UInt16        maj;
	UInt16        min;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	maj = sysGetROMVerMajor(romVersion);
	min = sysGetROMVerMinor(romVersion);
	if (maj > major)
		return true;
	else if ((maj == major) && (min >= minor))
		return true;
	
	return false;
}

static Boolean 
ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Int16    formId;
    Boolean handled = false;

    palmOS31 = calc_rom_greater(3,1);

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case frmMain:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	case frmItem:
	    FrmSetEventHandler(frm, ItemFormHandleEvent);
	    break;
	case frmPeople:
	    FrmSetEventHandler(frm, PeopleFormHandleEvent);
	    break;
	case frmDBmgr:
	    FrmSetEventHandler(frm, DBmgrFormHandleEvent);
	    break;
	case frmCurrency:
	    FrmSetEventHandler(frm, CurrencyFormHandleEvent);
	    break;
	case frmDescription:
	    FrmSetEventHandler(frm, DescriptionFormHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Err 
StartApplication(void)
{
    PrefData data;
    UInt16 size;
    
    dateFormat = (DateFormatType)PrefGetPreference(prefDateFormat);
    longDateFormat = (DateFormatType)PrefGetPreference(prefLongDateFormat);

    size = sizeof(data);
    if (PrefGetAppPreferences(APP_ID,PREF_DEFAULT,&data,
			      &size,false)!=PREF_VERSION) {
	FrmAlert(altAbout);
	FrmHelp(strHelp);
	FrmGotoForm(frmDBmgr);
	SysCopyStringResource(defaultCurrency,strDefCur);	
	activeBalance = false;
    } else {
	activeBalance = data.activeBalance;
	StrCopy(defaultCurrency,data.lcurrency);
	if (data.start == pref_main)
	    main_opendb(data.dbid,data.cardno);
	else
	    FrmGotoForm(frmDBmgr);
    }
    return 0;
}

/* Save preferences, close forms, close app database */
static void 
StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void 
EventLoop(void)
{
    Err err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

UInt32
PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
    Err err;

    if (launchCode == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) 
	    return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
