#!/usr/bin/env python

import prc
import os
import sys
import struct
import string
import time

class Amount:
    def __init__(self,value,currency):
        self.value = value
        self.currency = currency
        
    def __coerce__(self,other):
        if isinstance(other,Amount):
            return self,other
        return coerce(self.value,other)

    def __abs__(self):
        return Amount(abs(self.value),self.currency)

    def __neg__(self):
        return Amount(-self.value,self.currency)

    def __str__(self):
        return "%.2f%s" % (self.value,self.currency)

    def __add__(self,other):
        if type(other) == type(self) and self.currency == other.currency:
            return Amount(self.value+other.value,self.currency)
        raise RuntimeError("Cannot add items with differenc currency.")

    def __cmp__(self,other):
        if self.currency != other.currency:
            raise RuntimeError("Currencies do not agree!")
        return cmp(self.value,other.value)

class Currency:
    def __init__(self,raw=None,id=None,name=None,rate=None):
        if raw is not None:
            fmt = "!4sHd"
            name,self.id,self.rate = struct.unpack(fmt,
                                                   raw[:struct.calcsize(fmt)])
            self.name = name[:string.find(name,'\0')]
        else:
            self.name = name
            self.id = id
            self.rate = rate
            
    def __repr__(self):
        return "Currency(%s,%d,%f)" % (self.name,self.id,self.rate)
    def __str__(self):
        return self.name

class Person:
    def __init__(self,raw):
        fmt = "!Hd33s"
        self.id, coef, name = struct.unpack(fmt,raw[:struct.calcsize(fmt)])
        self.name = name[:string.find(name,'\0')]
    def __repr__(self):
        return "Person(%d,%s)" % (self.id,self.name)

class Item:
    def __init__(self,raw):
        fmt = "!HdHbxH33sxHxx"
        data = struct.unpack(fmt,raw[:struct.calcsize(fmt)])
        date,amount,payer,transfer,currency,description,npayees = data
        raw = raw[struct.calcsize(fmt):]

        fmt = "!Hd"
        pmts = []
        tsize = struct.calcsize(fmt)
        for i in range(npayees):
            pmts.append(struct.unpack(fmt,raw[:tsize]))
            raw = raw[tsize:]
            
        self.amount = amount
        self.payer = payer
        self.transfer = transfer
        self.currency = currency
        self.description = description[:string.find(description,'\0')]
        self.pmts = pmts
        self.date = (((date & 0xFE00) >> 9) + 1904,
                     (date & 0x01E0) >> 5,
                     date & 0x001F,
                     0,0,0,0,0,0)
    def __repr__(self):
        return "Item(%s,%.2f,%s)" % (self.description,self.amount,
                                   str(self.pmts))

class Database:
    PERSONTYPE = 1
    DESCRIPTIONTYPE = 2
    CURRENCYTYPE = 3
    ITEMTYPE = 5
    
    def __init__(self,pdbfile):
        self.people = {}
        self.items = []
        self.currency = {}
        self.indirect = 0
        self.curid = -1

        self.read_data(pdbfile)
        bname = os.path.basename(pdbfile)
        self.name = bname[4:-4]

    def set_curid(self,id):
        self.curid = id

    def person_name(self,id):
        return self.people[id]

    def read_data(self,pdbfile):
        pdb = prc.File(pdbfile,read=1,write=0)
        appdata = pdb.getAppBlock()

        # Support old databases
        try:
            lcurname,self.indirect = struct.unpack("!6x4shxx",appdata)
        except:
            lcurname,self.indirect = struct.unpack("!6x4sh",appdata)
            
        lcurname = lcurname[:string.find(lcurname,'\0')]

        self.currency[0] = Currency(name=lcurname,id=0,rate=1.0)
        # The sum functions automatically revers to '0', allow
        # easy currency naming in output
        self.currency[-1] = Currency(name=lcurname,id=0,rate=1.0)

        for i in range(pdb.getRecords()):
            raw,tmp,id,attr,category = pdb.getRecord(i)
            type = struct.unpack("!bx",raw[:2])[0]
            raw = raw[2:]
            if type == self.PERSONTYPE:
                person = Person(raw)
                self.people[person.id]=person.name
            elif type == self.CURRENCYTYPE:
                cur = Currency(raw=raw)
                self.currency[cur.id] = cur
            elif type == self.ITEMTYPE:
                self.items.append(Item(raw))
            elif type == self.DESCRIPTIONTYPE:
                pass
            else:
                print "BAD TYPE: ",type

    def calc_balances(self):
        """
        Calculate balances
        """
        result = {}
        for id in self.people.keys():
            result[id] = Amount(0.0,self.currency[self.curid].name)

        for id in self.people.keys():
            for item in self.items:
                result[id] = result[id] + self.calc_balance(id,item)

        return result

    def calc_payments(self):
        result = {}
        for id in self.people.keys():
            result[id] = Amount(0.0,self.currency[self.curid].name)

        for id in self.people.keys():
            for item in self.items:
                result[id] = result[id] + self.calc_payment(id,item)

        return result

    def calc_costs(self):
        """
        Calculate costs
        """
        if self.curid == -1:
            curid = 0
        else:
            curid = self.curid
        
        result = {}
        for id in self.people.keys():
            result[id] = Amount(0,self.currency[curid].name)

        for id in self.people.keys():
            for item in self.items:
                if item.transfer:
                    continue
                result[id] = result[id] + self.calc_cost(id,item,curid)

        return result

    def total_costs(self):
        """
        Calculate total cost of action in the given currency
        """
        result = 0.0
        for item in self.items:
            # Skip transfer items
            if item.transfer:
                continue
            tmpres = item.amount
            if item.currency != self.curid:
                tmpres = tmpres * self.currency[item.currency].rate
                tmpres = tmpres / self.currency[self.curid].rate
            result = result + tmpres
        return Amount(result,self.currency[self.curid].name)

    def calc_cost(self,person,item,curid=None):
        """
        Calculate the cost for person/item/currency
        """
        if curid is None and self.curid == -1:
            curid = item.currency
        elif curid is None:
            curid = self.curid
        
        result = 0.0
        sum = 0.0
        for coef in map(lambda x:x[1],item.pmts):
            sum = sum + coef
        for id,coef in item.pmts:
            if person == id:
                result = result + (item.amount/sum) * coef

        if item.currency != curid:
            result = result * self.currency[item.currency].rate
            result = result / self.currency[curid].rate
        return Amount(result,self.currency[curid].name)

    def calc_payment(self,person,item,curid=None):
        """
        Calculate payment for the current item and person and currency
        """
        result = 0.0

        if person == item.payer:
            result = result + item.amount

        if item.currency != self.curid:
            result = result * self.currency[item.currency].rate
            result = result / self.currency[self.curid].rate
        return Amount(result,self.currency[self.curid].name)

    def calc_balance(self,person,item):
        """
        Calculate balance for the current item and person and currency
        """
        result = 0.0
        sum = 0.0

        if person == item.payer:
            result = result + item.amount

        for coef in map(lambda x:x[1],item.pmts):
            sum = sum + coef
        for id,coef in item.pmts:
            if person == id:
                result = result - (item.amount/sum) * coef
        if item.currency != self.curid:
            result = result * self.currency[item.currency].rate
            result = result / self.currency[self.curid].rate
        return Amount(-result,self.currency[self.curid].name)

    def item_amount(self,item):
        """
        Return string repre of the total sum paid for item
        """
        result = item.amount
        if self.curid == -1:
            return Amount(result,self.currency[item.currency].name)

        if item.currency != self.curid:
            result = result * self.currency[item.currency].rate
            result = result / self.currency[self.curid].rate

        return Amount(result,self.currency[self.curid].name)

def html_escape(str):
    str = string.join(string.split(str, '&'), '&amp;')
    str = string.join(string.split(str, '<'), '&lt;')
    return str

def biggest(balances):
    max = 0
    mkey = None
    for key,val in balances.items():
        if abs(val) > max:
            mkey = key
            max = abs(val)
    return mkey

def biggest_plus(balances):
    max = 0
    mkey = None
    for key,val in balances.items():
        if val > max:
            mkey = key
            max = val
    return mkey

def biggest_minus(balances):
    max = 0
    mkey = None
    for key,val in balances.items():
        if -val > max:
            mkey = key
            max = -val
    return mkey

def calc_balance_procedure(balances):
    balances = balances.copy()
    result = []
    while 1:
        if len(balances) <= 1:
            return result
        mkey = biggest(balances)
        if mkey is None:
            return result
        
        amount = balances[mkey]
        del balances[mkey]
        
        while amount != 0:
            if amount > 0:
                nkey = biggest_minus(balances)
            else:
                nkey = biggest_plus(balances)
            if nkey is None:
                return result

            if abs(amount) > abs(balances[nkey]):
                pamount = -balances[nkey]
                amount = amount + balances[nkey]
                del balances[nkey]
            else:
                pamount = amount
                balances[nkey] = balances[nkey] + amount
                amount = 0

            if pamount > 0:
                result.append((mkey,nkey,pamount))
            else:
                result.append((nkey,mkey,-pamount))


class BillOutput:
    def __init__(self,o,db):
        self.o = o
        self.db = db
        self._linecounter = 0

    def get_color(self,ctype):
        shadeconst = 20
        if ctype == 'cost':
            c = (210,210,210)
        elif ctype == 'transfer':
            c = (150,230,230)
        elif ctype == '-cost':
            c = (230,170,170)
        elif ctype == 'summary':
            shadeconst = 0
            c = (170,170,170)
        elif ctype == 'surplus':
            shadeconst = 0
            c = (80,255,80)
        elif ctype == 'debtor':
            shadeconst = 0
            c = (255,80,80)
        
        self._linecounter += 1
        if self._linecounter % 2:
            c = [ x+shadeconst for x in c ]
        return '#%02x%02x%02x' % tuple(c)

    def generate(self):
        db = self.db
        self.write('<html><body>')
        self.heading(db.name)
        
        self.bill_start("whole group")
        lastdate = None
        for item in db.items:
            if lastdate != item.date:
                lastdate = item.date
                self.write_date(item.date)
            
            payer = db.person_name(item.payer)
            payees = [ db.person_name(x[0]) for x in item.pmts ]
            coefs = [ x[1] for x in item.pmts ]
            if item.transfer:
                self.whole_transfer(item.description,payer,payees,coefs,
                                    db.item_amount(item))
            else:
                self.whole_cost(item.description,payer,payees,coefs,
                                db.item_amount(item))
        self.summary('Total cost',self.db.total_costs())
        self.bill_end()

        total_costs = self.db.calc_costs()
        total_balances = self.db.calc_balances()
        total_payments = self.db.calc_payments()

        for pid,person in db.people.items():
            self.bill_start(person)
            lastdate = None
            for item in db.items:
                # Skip items that do not match the current person
                if not pid == item.payer and \
                       pid not in [ x[0] for x in item.pmts ]:
                    continue
                if lastdate != item.date:
                    lastdate = item.date
                    self.write_date(item.date)
                payer = db.person_name(item.payer)
                payees = [ db.person_name(x[0]) for x in item.pmts ]
                coefs = [ x[1] for x in item.pmts ]

                cost = db.calc_cost(pid,item)
                        
                if item.transfer:
                    self.person_transfer(item.description,payer,
                                         db.item_amount(item),
                                         cost,payees)
                else:
                    self.person_cost(item.description,payer,
                                     db.item_amount(item),
                                     cost,payees)
            self.summary('Total cost',total_costs[pid])
            self.summary('Total paid',total_payments[pid])
            if total_balances[pid] < 0.0:
                self.summary('Balance: you will get paid',-total_balances[pid],'surplus')
            else:
                self.summary('Balance: you will pay',total_balances[pid],'debtor')
            self.bill_end()
        self.write('</body></html>')
            
    def write(self,text):
        self.o.write(text)

    def summary(self,descr,amount,color='summary'):
        self.write('<tr bgcolor="%s">' % self.get_color(color))
        self.write('<th align="left" colspan="2">%s</th><td align="right">%s</td>\n' % \
                   (descr,amount))
        self.write('</tr>\n')

    def person_transfer(self,descr,payer,amount,cost,payees):
        self.write('<tr bgcolor="%s">\n' % self.get_color('transfer'))
        if cost == 0.0:
            self.write('<td>%s</td><td>Paid %s to: %s</td><td align="right">---</td>\n' %
                       (html_escape(descr),amount,", ".join(payees)))
        else:
            self.write('<td>%s</td><td>Received %s from %s (paid %s)</td><td align="right">---</td>\n' %
                       (html_escape(descr),cost,payer,amount))
        self.write('</tr>')
                
    def person_cost(self,descr,payer,amount,cost,payees):
        if amount > 0:
            action = 'paid'
            color = self.get_color('cost')
        else:
            color = self.get_color('-cost')
            action = 'received'
            amount = -amount
        
        self.write('<tr bgcolor="%s">\n' % color)
        if cost != 0.0:
            self.write('<td>%s</td><td>%s %s by %s</td><td align="right">%s</td>\n' %
                       (html_escape(descr),amount,action,payer,cost))
        else:
            self.write('<td>%s</td><td>%s %s %s for %s</td><td align="right">%s</td>\n' %
                       (html_escape(descr),payer,action,amount,
                        ", ".join(payees),cost))
        self.write('</tr>\n')
        
    def whole_transfer(self,descr,payer,payees,coefs,amount):
        pp = zip(payees,coefs)
        def format((name,coef)):
            if coef != 1.0:
                return "%s(%.1f)" % (name,coef)
            else:
                return name
            
        pp = ', '.join([format(x) for x in pp ])
        self.write('<tr bgcolor="%s">\n' % self.get_color('transfer'))
        self.write('<td>%s</td><td>Transfer from %s to %s</td><td align="right">%s</td>\n' %
                   (html_escape(descr),payer,pp,amount))
        self.write("</tr>\n")

    def whole_cost(self,descr,payer,payees,coefs,amount):
        if amount > 0:
            color = self.get_color('cost')
        else:
            color = self.get_color('-cost')
        
        pp = zip(payees,coefs)
        def format((name,coef)):
            if coef != 1.0:
                return "%s(%.1f)" % (name,coef)
            else:
                return name
            
        pp = ', '.join([format(x) for x in pp ])
        self.write('<tr bgcolor="%s">\n' % color)
        self.write('<td>%s</td><td>%s: %s</td><td align="right">%s</td>\n' % \
                   (html_escape(descr),payer,pp,amount))
        self.write('</tr>\n')

    def bill_start(self,text):
        self._linecounter = 0
        
        self.write('<table cellpadding="3" cellspacing="0" width="100%" border="0">\n')
        self.write('<tr bgcolor="#7777ff"><th colspan="3">Statement for %s</th></tr>\n' % text)
        self.write('<tr><th align="left">Description</th><th align="left">Info about payment</th><th align="right">Cost</th></tr>\n')

    def bill_end(self):
        self.write("</table><hr/>\n")

    def heading(self,name):
        self.write("<h1>Accounting Report of Action %s</h1>\n" % name)
        self.write("<h2>Date of Generation: %s</h2>\n" % \
                   time.strftime("%Y-%m-%d",time.localtime(time.time())))

    def write_date(self,date):
        self.write('<tr bgcolor="#c077ff"><td colspan="3" align="center">Date: %s</td></tr>\n' % \
                   time.strftime("%Y-%m-%d",date))
    

def gen_bill(o,db):
    gen = BillOutput(o,db)
    gen.generate() 

def gen_html(o,db):
    people = db.people
    currency = db.currency
    costs = [ x for x in db.items if not x.transfer ]
    transfers = [ x for x in db.items if x.transfer ]
    
    pids = people.keys()
    pids.sort()
    
    o.write('<table border="1" cellpadding="3" cellspacing="0">\n')
    # Header
    o.write('\n<thead>\n')
    o.write('<tr bgcolor="#99CCFF">\n')
    
    o.write('<th rowspan="2">Description</th><th rowspan="2">Date</th>')
    for id in pids:
        o.write('<th colspan=2>%s</th>' % html_escape(people[id]))
    
    o.write('</tr>\n')
    o.write('<tr bgcolor="#FF99FF">\n')
    for id in pids:
        o.write('<th>Credit</th><th>Debet</th>')
    o.write('</tr>')
    o.write('</thead>\n')

    # Costs
    o.write('<tbody>')
    o.write('<tr bgcolor="#99FF99"><td align="center" colspan="%d">COSTS</td></tr>\n' % (2+2*len(pids)))
    for cost in costs:
        o.write('<tr>')
        o.write('<td>%s</td>' % html_escape(cost.description))
        o.write('<td>%s</td>' % time.strftime("%Y-%m-%d",cost.date))
        for id in pids:
            if id == cost.payer:
                o.write('<td bgcolor="#CC99FF">%s</td>' % db.item_amount(cost))
            else:
                o.write('<td></td>')
            if id in map(lambda x:x[0],cost.pmts):
                o.write('<td bgcolor="#FF9999">%s</td>' % \
                        db.calc_cost(id,cost))
            else:
                o.write('<td></td>')
        o.write('</tr>\n')
    o.write('</tbody>')

    # Transfers
    o.write('<tbody>')
    o.write('<tr bgcolor="#99FF99"><td align="center" colspan="%d">TRANSFERS</td></tr>\n' % (2+2*len(pids)))
    for cost in transfers:
        o.write('<tr>')
        o.write('<td>%s</td>' % html_escape(cost.description))
        o.write('<td>%s</td>' % time.strftime("%Y-%m-%d",cost.date))
        for id in pids:
            if id == cost.payer:
                o.write('<td bgcolor="#CC99FF">%s</td>' % db.item_amount(cost))
            else:
                o.write('<td></td>')
            if id in map(lambda x:x[0],cost.pmts):
                o.write('<td bgcolor="#FF9999">%s</td>' % \
                        db.calc_cost(id,cost))
            else:
                o.write('<td></td>')
        o.write('</tr>\n')
    o.write('</tbody>\n')
    # Summary
    o.write('<tbody>')
    o.write('<tr bgcolor="#99FF99"><td align="center" colspan="%d">SUMMARY</td></tr>\n' % (2+2*len(pids)))
    # List of people again
    o.write('<tr bgcolor="#99CCFF">\n')
    o.write('<th></th><th></th>')
    for id in pids:
        o.write('<th colspan=2>%s</th>' % html_escape(people[id]))
    
    o.write('</tr>\n')
    
    # Costs
    o.write('<tr>')
    o.write('<td>Total cost</td><td>%s</td>' % \
            time.strftime("%Y-%m-%d",time.localtime(time.time())))
    costres = db.calc_costs()
    for id in pids:
        o.write('<td colspan="2" align="center" bgcolor="#FF9999">%s</td>\n' % \
                costres[id])
    o.write('</tr>')
    o.write('</tbody>')

    # Balances
    o.write('<tbody>')
    o.write('<tr>')
    o.write('<td>Balance</td><td>%s</td>' % \
            time.strftime("%Y-%m-%d",time.localtime(time.time())))
    balances = db.calc_balances()
    for id in pids:
        if balances[id] <= 0.0:
            o.write('<td colspan="2" align="center" bgcolor="#33FF33">%s</td>' %\
                    balances[id])
        elif balances[id] > 0.0:
            o.write('<td colspan="2" align="center" bgcolor="#FF3333">%s</td>' %\
                    balances[id])
    
    o.write('</tr>')
    o.write('</tbody>')

    # Total cost
    o.write('<tr bgcolor="#99FF99"><td align="center" colspan="%d">TOTAL COST: %s</td></tr>\n' % ((2+2*len(pids)),db.total_costs()))
    o.write('</table>\n')


    if (len(currency) > 2):
        curid = db.curid
        if curid == -1:
           curid = 0
           
        # Currency information 
        o.write('<table border="1" cellpadding="3" cellspacing="0">\n')
        # Header
        o.write('\n<thead>\n')
        o.write('<tr bgcolor="#99FF99">\n')
        o.write('<td colspan="%d">Currency exchange rates</td>' % (len(currency)-2))
        o.write('</tr>')
        
        o.write('<tr>\n')
        for id in currency.keys():
            if id == -1 or id == curid:
                continue
            o.write('<th bgcolor="#FF99FF">%s/%s</th>' % (currency[curid],
                                                          currency[id]))
        o.write('</tr>')
        o.write('</thead>\n')
        
        o.write('<tbody><tr>')
        for id in currency.keys():
            if id == -1 or id == curid:
                continue
            o.write('<td bgcolor="#FF9999">%.4f</td>' % (currency[id].rate/currency[curid].rate))
        o.write('</tr></tbody>')
        
        o.write('</table>')


    # Proposed balance procedure

    o.write('<table border=0>')
    o.write('<tr bgcolor="#99FF99"><th align="center" colspan=4>Proposed balance procedure</th></tr>')
    balances = db.calc_balances()
    procbal = calc_balance_procedure(balances)
    for payer,payee,amount in procbal:
        o.write('<tr><td>%s</td><td>------&gt;</td><td>%s</td><td>%s</td></tr>' %
                (html_escape(people[payer]),
                 html_escape(people[payee]),
                 amount))

    o.write('</table>')

def gen_csv(o,db):
    people = db.people
    currency = db.currency
    costs = [ x for x in db.items if not x.transfer ]
    transfers = [ x for x in db.items if x.transfer ]
    
    o.write('"Description","Date"')
    pids = people.keys()
    pids.sort()
    for pid in pids:
        o.write(',"CR: %s","DB: %s"' % (people[pid],people[pid]))
    o.write('\n')

    o.write('"Costs"\n')
    for cost in costs:
        o.write('"%s",%s' % (cost.description,
                               time.strftime("%Y-%m-%d",cost.date)))
        for id in pids:
            if id == cost.payer:
                o.write(',%s' % db.item_amount(cost))
            else:
                o.write(',0')
            if id in map(lambda x:x[0],cost.pmts):
                o.write(',%s' % \
                        db.calc_cost(id,cost))
            else:
                o.write(',0')
        o.write('\n')

    o.write('"Transfers"\n')
    for cost in transfers:
        o.write('"%s","%s"' % (cost.description,
                               time.strftime("%Y-%m-%d",cost.date)))
        for id in pids:
            if id == cost.payer:
                o.write(',%s' % db.item_amount(cost))
            else:
                o.write(',0')
            if id in map(lambda x:x[0],cost.pmts):
                o.write(',%s' % \
                        db.calc_cost(id,cost))
            else:
                o.write(',0')
        o.write('\n')

    o.write('"Summary"\n')
    o.write('"Balances",%s' % \
            time.strftime("%Y-%m-%d",time.localtime(time.time())))
    balances = db.calc_balances()
    for id in pids:
        o.write(',%s,' % balances[id])
    o.write('\n')
    
    costres = db.calc_costs()
    o.write('"Costs",')
    for id in pids:
        o.write(',%s,' %  costres[id])
    o.write('\n')
    

def main(argv):
    global currency
    
    if (len(argv) != 3 and len(argv) != 4) or argv[1] not in ('html','csv','bill'):
        sys.stderr.write("Usage: %s {html|bill|csv} [currency] <database.pdb>\n" % argv[0])
        sys.exit(1)

    if len(argv) == 3:
        pdbfile = argv[2]
        curid = -1
    else:
        pdbfile = argv[3]
        curname = argv[2]
        
    db = Database(pdbfile)

    if len(argv) == 4:
        for cur in db.currency.values():
            if cur.name == curname:
                curid = cur.id
                break
        else:
            sys.stderr.write("Incorrect currency name.\n")
            sys.exit(1)


    db.set_curid(curid)

    if argv[1] == 'csv':
        gen_csv(sys.stdout,db)
    elif argv[1] == 'bill':
        gen_bill(sys.stdout,db)
    elif argv[1] == 'html':
        sys.stdout.write('<html><body>\n')
        gen_html(sys.stdout,db)
        sys.stdout.write('</body></html>')

if __name__ == '__main__':
    main(sys.argv)

